# ==
# --Project: Open Grid
# --Author: Sing.Yu <scaner@gmail.com>
# --Module: og.lib.async
# ==

# --
# This file is referenced from Python's asyncore.py
# --


# ==
# Internal Var name table
# _ch - Poller
# socket - Core socket.
# wreg - Write event flag
# rreg - Read event flag
# ==

# ==
# DispatcherBase events:
#  - onEventRead
#  - onEventWrite
#  - onEventException
#  - onEventError
# ==

# ==
# SocketDispatcher events:
#  - onRead
#  - onWrite
#  - onConnect
#  - onClose
#  - onException
#  - onAccept
# ==

# ==
# OGConnectio's handler event:
#  - setConnection
#  - onMessageIn
#  - onMessageOut
#  - onConnect
#  - onClose
# ==

# == == == == == == == == == == == == == == == == == == == == == == #
import sys, os
import time
import socket
import select

from errno import \
	EALREADY, EINPROGRESS, EWOULDBLOCK, ECONNRESET, \
	ENOTCONN, ESHUTDOWN, EINTR, EISCONN, errorcode

# ---
from og.lib.utility import CompactTrackback
from og.lib.network import Str2Dword, Unpack_ as OGUnpack, Pack as OGPack

# == == == == == == == == == == == == == == == == == == == == == == #
class ExitNow(Exception): pass

RFLAGS = select.POLLIN | select.POLLPRI
WFLAGS = select.POLLOUT
EFLAGS = select.POLLERR | select.POLLHUP | select.POLLNVAL

# ---
def _log(msg):
	sys.stderr.write('log: %s\n' % msg)

# ---
def ReadWrite(o, flags):
	try:
		if flags & RFLAGS:
			o.onEventRead()
		if flags & WFLAGS:
			o.onEventWrite()
		if flags & EFLAGS:
			o.onEventException()

	except ExitNow:
		raise
	except:
		o.onEventError()
		

# -- Poller
class Poller(object):
	# -- Poller.__init__
	def __init__(self):
		self.map = {}
		self.poller = select.poll()
		self.alive = True
		self.check = None

	def register(self, fd, emask):
		self.poller.register(fd, emask)

	def unregister(self, fd):
		self.poller.unregister(fd)
	
	def stop(self):
		self.alive = False
		
	def mainloop(self):
		map = self.map
		poller = self.poller
				
		while self.alive:
			try:
				r = poller.poll(500)
			except select.error, err:
				if err[0] != EINTR:
					raise
				r = []
			for fd, flags in r:
				o = map.get(fd)
				if o is None:
					continue
				ReadWrite(o, flags)
			
			self.check()
# -- End of Poller
	

# -- DispatcherBase
class DispatcherBase(object):
	# -- DispatcherBase.__init__
	def __init__(self, ch):
		self._ch = ch
		self.rreg = False
		self.wreg = False

	# -- DispatcherBase.attachChannel
	def attachChannel(self):
		self._ch.map[self._fileno] = self
	
	# -- DispatcherBase.detachChannel
	def detachChannel(self):
		map = self._ch.map
		if map.has_key(self._fileno):
			del map[self._fileno]
		self._ch = None
	
	# -- DispatcherBase.registerRead
	def registerRead(self, flags):
		if self.rreg == flags:
			return
		self.rreg = flags
		self.register()

	# -- DispatcherBase.registerWrite
	def registerWrite(self, flags):
		if self.wreg == flags:
			return
		self.wreg = flags
		self.register()
	
	# -- DispatcherBase.register
	def register(self):
		emask = 0
		if self.wreg:
			emask |= select.POLLOUT
		if self.rreg:
			emask |= select.POLLIN
		
		if emask:
			self._ch.register(self._fileno, emask)
		else:
			self._ch.unregister(self._fileno)

	# -- DispatcherBase.onEventError
	def onEventError(self):
		nil, t, v, tbinfo = CompactTrackback()
		
		self_repr = repr(self)
		_log('uncaptured python exception, closing channel %s\n%s\n%s\n%s' % (
			self_repr, t, v, tbinfo))
		#_log('unhandled error event')
# -- End of DispatcherBase
	

# == == == == == == == == == == == == == == == == == == == == == == #
# -- SocketDispatcher
class SocketDispatcher(DispatcherBase):
	connected = False
	accepting = False
	closed = False

	# -- SocketDispatcher.__init__
	def __init__(self, ch, sock = None):
		DispatcherBase.__init__(self, ch)
		if sock:
			self.setSocket(sock)
			self.socket.setblocking(0)
			self.connected = True
		else:
			self.socket = None
	
	# -- SocketDispatcher.createSocket
	def createSocket(self, family, type):
		self.socket = socket.socket(family, type)
		self.socket.setblocking(0)
		self._fileno = self.socket.fileno()
		self.attachChannel()
	
	# -- SocketDispatcher.createTCPSocket
	def createTCPSocket(self):
		self.createSocket(socket.AF_INET, socket.SOCK_STREAM)
	
	# -- SocketDispatcher.setSocket
	def setSocket(self, sock):
		self.socket = sock
		self._fileno = sock.fileno()
		self.attachChannel()
	
	# -- SocketDispatcher.setReuseAddr
	def setReuseAddr(self):
		try:
			self.socket.setsockopt(
				socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
		except socket.error:
			pass
	
	# -- SocketDispatcher.bind
	def bind(self, addr):
		return self.socket.bind(addr)

	# -- SocketDispatcher.listen
	def listen(self, num):
		self.accepting = True
		return self.socket.listen(num)
	
	# -- SocketDispatcher.connect
	def connect(self, addr):
		self.connected = False
		err = self.socket.connect_ex(addr)

		if err in (EINPROGRESS, EALREADY, EWOULDBLOCK):
			self.registerWrite(True)
			return
		if err in (0, EISCONN):
			self.connected = True
			self.onConnect()
		else:
			raise socket.error, (err, errorcode[err])

	
	# -- SocketDispatcher.accept
	def accept(self):
		try:
			conn, addr = self.socket.accept()
			return conn, addr
		except socket.error, why:
			if why[0] == EWOULDBLOCK:
				pass
			else:
				raise
	
	# -- SocketDispatcher.send
	def send(self, data):
		try:
			result = self.socket.send(data)
			return result
		except socket.error, why:
			if why[0] == EWOULDBLOCK:
				return 0
			else:
				raise
			return 0
	
	# ---
	def recvfrom(self, bs):
		try:
			return self.socket.recvfrom(bs)
		except socket.error, why:
			if why[0] == EWOULDBLOCK:
				return None, None
			raise

	# -- SocketDispatcher.recv
	def recv(self, bs):
		try:
			data = self.socket.recv(bs)
			if not data:
				self.onClose()
				return ''
			else:
				return data
		except socket.error, why:
			if why[0] == EWOULDBLOCK:
				return ''
			if why[0] in [ECONNRESET, ENOTCONN, ESHUTDOWN]:
				self.onClose()
				return ''
			else:
				raise

	# -- SocketDispatcher.close
	def close(self):
		if self.closed:
			return
		self.registerRead(False)
		self.registerWrite(False)
		self.detachChannel()
		self.socket.close()
		self.closed = True
	
	# -- SocketDispatcher.onEventRead
	def onEventRead(self):
		if self.accepting:
			if not self.connected:
				self.connected = True
			self.onAccept()
		elif not self.connected:
			self.onConnect()
			self.connected = True
			self.onRead()
		else:
			self.onRead()

	def onEventWrite(self):
		if not self.connected:
			self.onConnect()
			self.connected = True
		self.onWrite()
	
	def onEventException(self):
		self.onException()
	
	
	def onAccept(self):
		_log('unhandled accept event')

	# -- SocketDispatcher.onException
	def onException(self):
		#_log('unhandled exception event')
		self.onClose()

	def onRead(self):
		_log('unhandled read event')
	def onWrite(self):
		_log('unhandled write event')
	def onConnect(self):
		_log('unhandled connect event')
	def onClose(self):
		_log('unhandled close event')
		self.close()
# --- End of SocketDispatcher	

class FDDispatcher(DispatcherBase): pass

# -- TCPServerDispatcher
class TCPServerDispatcher(SocketDispatcher):
	# -- TCPServerDispatcher.__init__
	def __init__(self, ch):
		SocketDispatcher.__init__(self, ch)

	# -- TCPServerDispatcher.initialize
	def initialize(self, addr):
		self.createSocket(socket.AF_INET, socket.SOCK_STREAM)
		self.setReuseAddr()
		self.socket.bind(addr)
		self.listen(100)
		self.registerRead(True)
# -- End of TCPServerDispatcher

# -- TCPIncoming
class TCPIncoming(TCPServerDispatcher):
	def __init__(self, parent, linkClass):
		TCPServerDispatcher.__init__(self, parent.poller)
		self.parent = parent
		self.linkClass = linkClass
	
	def onAccept(self):
		s, addr = self.accept()
		s.setblocking(0)

		C = OGConnection(self.parent.poller)
		C.setSocket(s)
		H = self.linkClass(self.parent)
		C.setHandler(H)
		C.enable()

# == == == == == == == == == == == == == == == == == == == == == == #
OG_MESSAGE_HEADERLEN = 4
OG_STATE_HEAD = 0
OG_STATE_BODY = 1

# -- OGConnection
class OGConnection(SocketDispatcher):
	# -- OGConnection.__init__
	def __init__(self, ch):
		SocketDispatcher.__init__(self, ch)
		self.inBuffer = []
		self.outBuffer = []
		self.inWant = 0
		self.inGot = 0
		self.state = 0
	
	# -- OGConnection.enable
	def enable(self):
		self.registerRead(True)
		
	# -- OGConnection.setHandler
	def setHandler(self, handler):
		handler.setConnection(self)
		self.handler = handler
		
	# -- OGConnection.onConnect
	def onConnect(self):
		self.handler.onConnect()
		self.inWant = OG_MESSAGE_HEADERLEN
		self.state = OG_STATE_HEAD
		self.registerRead(True)
	
	# -- OGConnection.onClose
	def onClose(self):
		self.handler.onClose()
		SocketDispatcher.close(self)
	
	# -- OGConnection.onRead
	def onRead(self):
		while not self.closed:
			buf = self.recv(self.inWant - self.inGot)
					
			if not buf:
				return
			
			self.inGot += len(buf)
			self.inBuffer.append(buf)
			
			if self.inGot != self.inWant:
				continue
			buf = ''.join(self.inBuffer)
			if self.state == OG_STATE_HEAD:
				self.inGot = 0
				self.inWant = Str2Dword(buf)
				self.inBuffer = []
				self.state = OG_STATE_BODY
			else:
				self.inGot = 0
				self.inWant = OG_MESSAGE_HEADERLEN
				self.inBuffer = []
				self.state = OG_STATE_HEAD
				
				inMessage = OGUnpack(buf)
				self.wreg2 = self.wreg
				self.handler.onMessageIn(inMessage)

				while not self.wreg2 and self.wreg:
					self.flushOut()

	# -- OGConnection.onWrite
	def onWrite(self):
		self.flushOut()

	# -- OGConnection.flushOut
	def flushOut(self):
		if not self.outBuffer:
			self.registerWrite(False)
			return
		
		while self.outBuffer and not self.closed:
			b0 = self.outBuffer[0]
			o = self.send(b0)
			if o != len(b0):
				if o != 0:
					self.outBuffer[0] = b0[o:]
				return	
			self.outBuffer.pop(0)


		if not self.outBuffer:
			self.wreg2 = False
			self.registerWrite(False)
			self.handler.onMessageOut()
		else:
			self.wreg2 = True

	# -- OGConnection.sendBuffer
	def sendBuffer(self, buf):
		self.outBuffer.append(buf)
		self.registerWrite(True)
		
	# -- OGConnection.sendMessage
	def sendMessage(self, params):
		self.sendBuffer(OGPack(params))
		
	# -- OGConnection.sendMessage0
	def sendMessage0(self, *params):
		self.sendBuffer(OGPack(params))

# -- End of OGConnection

# == == == == == == == == == == == == == == == == == == == == == == #
class MessageChannel(object):
	# -- MessageChannel.__init__
	def __init__(self):
		self.C = None # handler
		self.oMessage = None
		self.oMessages = []
		self.iMessage = None

	# -- MessageChannel.messageOut
	def messageOut(self, msg):
		self.oMessages.append(msg)
		self.pushOutMessages()
			
	# -- MessageChannel.onMessageOut
	def onMessageOut(self):
		self.oMessage = None
		self.pushOutMessages()

	# -- MessageChannel.setConnection
	def setConnection(self, c):
		self.C = c
	
	# -- MessageChannel.pushOutMessages
	def pushOutMessages(self):
		if not self.C: return
		if self.oMessage: return
		if not self.oMessages: return
		self.oMessage = self.oMessages.pop(0)
		self.C.sendMessage(self.oMessage)
	
	# -- MessageChannel.close
	def close(self):
		self.C.onClose()

	# --
	def onConnect(self): pass
	def onClose(self): pass
	def onMessageIn(self, msg): pass
