# ==
# --Project: Open Grid
# --Author: Sing.Yu <scaner@gmail.com>
# --Module: og.lib.osmeter.linux
# ==

__all__ = ['Meter']

# --
import os
import time
import socket

# --
from meter import Slurp, SlurpLines, Partition

# --
class LinuxMeter:
	# -- LinuxMeter.__init__
	def __init__(self):
		self.uname = os.uname()
		self.hostname = socket.gethostname()

		self.cpuinfo = {}
		self.cpu()

		self.meminfo = {}
		self.mem()
		self.totalmem = '%dG' % int(self.meminfo['MemTotal'] / 1024.0 / 1024 + 0.5)

		self.diskinfo = {}
		self.disk()

		self.netinfo = {}
		self.net()

		self.sysinfo = {}
		self.sys()

	# -- LinuxMeter.cpu	
	def cpu(self):
		lines = SlurpLines('/proc/cpuinfo')

		phyids = set()
		processors = 0

		for line in lines:
			n, v = Partition(line)
			if n == 'processor':
				processors += 1
			if n == 'physical id':
				phyids.add(v)
			if n == 'model name':
				modelname = v
			if n == 'flags':
				flags = v

		self.cpuinfo['modelname'] = modelname
		self.cpuinfo['processors'] = processors
		self.cpuinfo['flags'] = flags

		if len(phyids) != 0 and len(phyids) != processors:
			self.cpuinfo['ht'] = 2
		else:
			self.cpuinfo['ht'] = 1
		self.cpuinfo['cpus'] = processors / self.cpuinfo['ht']



	# -- LinuxMeter.mem
	def mem(self):
		lines = SlurpLines('/proc/meminfo')
		for line in lines:
			n, v = Partition(line)
			if n in ('MemTotal', 'MemFree', 'Buffers', 
				'Cached', 'SwapTotal', 'SwapFree'):
				self.meminfo[n] = int(v[:-3])
	

	# -- LinuxMeter.disk
	def disk(self):
		lines = SlurpLines('/etc/mtab')

		for line in lines:
			words = line.split()
			point = words[1]
			fstype = words[2]

			if fstype not in ('ext2', 'ext3'):
				continue
			self.diskinfo[point] = os.statvfs(point)
			
	# -- LinuxMeter.net
	def net(self):
		cur = {}
		lines = SlurpLines('/proc/net/dev')
		for line in lines[1:]:
			n, v = Partition(line)
			if n.startswith('eth'):
				values = [ int(x) for x in v.split() ]
				inbyte = values[0]
				inpacket = values[1]
				outbyte = values[8]
				outpacket = values[9]
				cur[n] = (inbyte, inpacket, outbyte, outpacket)
				cur['cur'] = time.time()
		self.netinfo['last'] = self.netinfo.get('cur')
		self.netinfo['cur'] = cur
	
	# -- LinuxMeter.netSpeed
	def netSpeed(self):
		cur = self.netinfo.get('cur')
		last = self.netinfo.get('last')
		speed = {}
		if not cur or not last:
			return

		during = cur['cur'] - last['cur']
		speed['during'] = during

		for dev in cur.keys():
			if dev == 'cur':
				continue

			if dev not in last:
				continue

			curdev = cur[dev]
			lastdev = last[dev]
			speeddev = []
			for i in range(4):
				cdata = curdev[i]
				ldata = lastdev[i]
				if cdata > ldata:
					speeddev.append(int((cdata - ldata) / during))
				else:
					speeddev.append(int((cdata + 0x100000000 - ldata) / during))
			speed[dev] = speeddev

		self.netinfo['speed'] = speed
				
	# -- LinuxMeter.sys
	def sys(self):
		# average load
		# http://www.linuxinsight.com/proc_loadavg.html
		line = Slurp('/proc/loadavg').strip()
		self.sysinfo['loadavg'] = line.split()
	
		# uptime
		# http://www.linuxinsight.com/proc_uptime.html
		line = Slurp('/proc/uptime').strip()
		self.sysinfo['uptime'] = line.split()

		# stat
		# http://www.linuxhowtos.org/System/procstat.htm
		lines = SlurpLines('/proc/stat')
		for line in lines:
			line.strip()
			tokens = line.split()
			if tokens[0].startswith('cpu'):
				self.sysinfo[tokens[0]] = [int(x) for x in tokens[1:]]

	# -- LinuxMeter.touch
	def touch(self):
		self.net()
		self.netSpeed()
		self.disk()
		self.mem()
		self.sys()

	# -- LinuxMeter.info
	def info(self):
		res = {
			'net': self.netinfo,
			'sys': self.sysinfo,
			'cpu': self.cpuinfo,
			'mem': self.meminfo,
			'disk': {}
		}

		for k,v in self.diskinfo.items():
			res['disk'][k] = list(v)
		return res
			
# -- End Of LineMeter
Meter = LinuxMeter
