# ==
# --Project: Open Grid
# --Author: Sing.Yu <scaner@gmail.com>
# --Module: og.lib.process
# ==

import os, sys

from og.lib.logging import *

# ---
try:
	MAXFD = os.sysconf('SC_OPEN_MAX')
except (AttributeError, ValueError):
	MAXFD = 256

# ---
# Flags:
#   N: normal, nouse flag, default value
#   P: Stdin/Stdout/Stderr in Pipe
#   r: repeat
# ---

# ---
# Notifier:
#   onFork(PI, pid)
#   onExecute(PI)
#   onReap(PI, pid, status)
# ---

# == == == == == == == == == == == == == == == == == == == == == == #
class PINotiferBase:
	def onFork(self, PI, pid): pass
	def onExecute(self, PI): pass
	def onReap(self, PI, pid, status): pass

# ---
class ProcInfo:
	# -- ProcInfo.__init__
	def __init__(self, manager, **P):
		self.manager = manager

		self.name = P.get('name')
		self.args = P.get('args')
		self.env = P.get('env')
		self.flags = P.get('flags', 'N')
		self.info = P.get('info', {})
		self.pid = 0

		self.stdin = None
		self.stdout = None
		self.stderr = None
		
		self.channel = None
		self.notifier = None
	
	# -- ProcInfo.setNotifier
	def setNotifier(self, notifier):
		self.notifier = notifier

	# -- ProcInfo.fork1
	def fork1(self):
		pid = os.fork()
	
		if pid != 0:
			# Main
			self.manager.procs[pid] = self
			if self.notifier: self.notifier.onFork(self, pid)
			return pid

		if self.notifier: self.notifier.onFork(self, pid)
		for fd in range(3, MAXFD):
			try:
				os.close(fd)
			except OSError:
				pass

		self.execute()
		
	
	# -- ProcInfo.fork2
	def fork2(self):
		r0, w0 = os.pipe()
		r1, w1 = os.pipe()
		r2, w2 = os.pipe()

		self.stdin = w0
		self.stdout = r1
		self.stderr = r2

		pid = os.fork()

		if pid != 0:
			os.close(r0)
			os.close(w1)
			os.close(w2)

			self.manager.procs[pid] = self
			if self.notifier: self.notifier.onFork(self, pid)
			return pid

		for i in range(0, 3):
			try:
				os.close(i)
			except OSError:
				pass

		os.dup2(r0, 0)
		os.dup2(w1, 1)
		os.dup2(w2, 2)

		for fd in range(3, MAXFD):
			try:
				os.close(fd)
			except OSError:
				pass
		self.execute()


	# -- ProcInfo.execute
	def execute(self):
		if self.notifier:
			self.notifier.onExecute(self)

		if self.env:
			os.environ.update(self.env)

		rcode = os.execvpe(self.args[0], self.args, os.environ)
		os.exit(rcode)
		
	# -- ProcInfo.fire
	def fire(self):
		if self.flags.find('P') != -1:
			self.pid = self.fork2()
		else:
			self.pid = self.fork1()

		_D('PI.fire %s: %s -- %d' % (
			self.name, ' '.join(self.args), self.pid))
		return self.pid

	# -- ProcInfo.reap
	def reap(self, pid, status):
		_D('PI.dead %s -- %d' % (self.name, self.pid))
		if self.notifier:
			self.notifier.onReap(self, pid, status)

		if self.flags.find('r') != -1:
			self.fire()

PI = ProcInfo
# -- end of ProcInfo


# == == == == == == == == == == == == == == == == == == == == == == #
class ProcManager:
	# -- ProcManager.__init__
	def __init__(self, managing):
		self.managing = managing
		self.procs = {}
	
	# -- ProcManager.reapZombie
	def reapZombie(self):
		while True:
			try:
				pid, status = os.waitpid(0, os.WNOHANG)
			except OSError, why:
				if why[0] == 10: # ECHILD
					break
				raise

			if pid == 0:
				break

			process = self.procs.get(pid)
			if process:
				del self.procs[pid]
				process.reap(pid, status)
	
	# -- ProcManager.killall
	def killall(self):
		for pid in self.procs.keys():
			os.kill(pid)
		self.reapZombie()
# -- End of ProcManager
PM = ProcManager
