# ==
# --Project: Open Grid
# --Author: Sing.Yu <scaner@gmail.com>
# --Module: og.svc.farming
# ==

FARMING_VERSION = '0.1-r8'

# ==
# Class list:
#   Farming
#   FarmIncoming
#   FarmLink - Incoming request link
#   Grid
#   GridLink - Outgoing link to Open Grid hiving service
#   FarmTask
# ==

# ==
# Farming Service:
#   Open Grid Management Gateway
# ==

import os, sys
import signal
import time
import socket

from og.lib.logging import *
from og.lib.logging import logCore

import og.constants
from og.lib.async import Poller, TCPIncoming
from og.lib.async import OGConnection, MessageChannel
from og.lib.utility import SecondTimer, ListenAddr, LoadConfig3
from og.lib.utility import Serial, MessageReply

# ---- INTERVAL
INTERVAL_GRIDS = 2
INTERVAL_UPDATE_GRIDINFO = 60

# ---
TaskMap = {}

# ----
# Farming - Farming Service root class
class Farming(object):
	def __init__(self):
		self.poller = Poller()
		self.poller.check = self.secondCheck
		self.alive = True
	
		self.nodeid = 'farm'

	# -- Farming.setup
	# Setup farming service.
	def setup(self, C):
		self.config = C
		self.myconfig = C['farm']
		
		self.ss = TCPIncoming(self, FarmLink)
		self.ss.initialize(ListenAddr(None, og.constants.FARMING_PORT))
		UpdateCurrent()
		self.lastCheck = Current.I
		self.startid = Current.I

		self.grids = {}
		for gridinfo in self.myconfig['grids'].values():
			grid = Grid(self)
			grid.setup(gridinfo)
			self.grids[grid.name] = grid

		self.timer = SecondTimer(self)
		self.timer.addCallback('grids', Farming.checkGrids, INTERVAL_GRIDS)

	# == Farming.checks
	# -- Farming.onMessageIn(self, source, msg):
	def onMessageIn(self, source, msg):
		taskid = msg['!'].get('=')
		if taskid is None:
			return
		task = TaskMap.get(taskid)

		if task is None:
			return
		if task.finished:
			return

		task.onMessageIn(msg)
		
	# -- Farming.secondCheck
	def secondCheck(self):
		UpdateCurrent()
		curI = Current.I

		if curI == self.lastCheck:
			return

		self.lastCheck = curI
		self.timer.touch(curI)
		self.cleanTaskMap(curI)
	
	# -- Farming.cleanTaskMap
	def cleanTaskMap(self, curI):
		keys = TaskMap.keys()
		for key in keys:
			tasklet = TaskMap[key]
			if curI - tasklet.last > tasklet.timeout:
				tasklet.onTimeout()
		
	# -- Farming.checkGrids
	def checkGrids(self, curI):
		for grid in self.grids.keys():
			self.grids[grid].touch(curI)
	
	# -- Farming.run
	def run(self):
		_I('Farming-%s:%d ready' % (FARMING_VERSION, os.getpid()))
		self.poller.mainloop()
# -- End of Farming

# -- Grid
class Grid(object):
	# -- Grid.__init__
	def __init__(self, parent):
		self.parent = parent
		self.link = GridLink(self)
		self.connected = False

	# -- Grid.setup
	def setup(self, config):
		self.addr = config['addr']
		self.name = config['name']
		self.domain = config['domain']
		self.info = dict(config)
		self.link = GridLink(self)
		self.last = {
			'gridinfo': Current.I
		}
	
	# -- Grid.touch
	def touch(self, curI):
		if self.link.state == 'closed':
			self.link.connect()
		if self.link.state != 'closed' and curI - self.last['gridinfo'] > INTERVAL_UPDATE_GRIDINFO:
			self.last['gridinfo'] = curI
			GridUpdateInfoTasklet(self).fire()
	
# -- GridLink
class GridLink(MessageChannel):
	# -- GridLink.__init__
	def __init__(self, parent):
		MessageChannel.__init__(self)
		self.parent = parent # Grid
		self.root = parent.parent # Farming
		self.C = None
		self.state = 'closed'
	
	# -- GridLink.isActive
	def isActive(self):
		return self.C != None

	# -- GridLink.connect
	def connect(self):
		self.state = 'connecting'
		c =  OGConnection(self.root.poller)
		c.createTCPSocket()
		c.setHandler(self)
		self.C.connect(self.parent.addr)
	
	# -- GridLink.onConnect
	def onConnect(self):
		self.state = 'connected'
		startup = GridStartupUpdateInfoTasklet(self.parent)
		startup.fire()

	# -- GridLink.onMessageIn
	def onMessageIn(self, msg):
		self.root.onMessageIn(self, msg)
	
	# -- GridLink.onClose
	def onClose(self):
		self.state = 'closed'
		self.C = None
		if self.oMessage:
			self.oMessages.insert(0, self.oMessage)
			self.oMessage = None
# -- End of Grid Link


# -- FarmLink
class FarmLink(MessageChannel):
	# -- FarmLink.__init__
	def __init__(self, parent):
		MessageChannel.__init__(self)
		self.parent = parent
		self.id = Serial()
		self.tasklet = None
	
	# -- FarmLink.replyMessageWithResult
	def replyMessageWithResult(self, msg, res):
		remsg = MessageReply(msg)
		remsg['='] = res
		self.messageOut(remsg)
	
	# -- FarmLink.onMessageIn
	def onMessageIn(self, msg):
		if '!' in msg and '=' in msg['!']:
			return self.parent.routing(self, msg)

		command = msg['_'][2:]
		func = getattr(FarmLink, 'message__' + command)
		if func:
			func(self, msg)
	
	# -- FarmLink.message__flushinfo
	def message__flushinfo(self, msg):
		res = []

		for grid in self.parent.grids.values():
			if grid.link.state != 'closed':
				GridUpdateInfoTasklet(grid).fire()
				grid.last['gridinfo'] = Current.I 
				res.append(grid.name)

		self.replyMessageWithResult(msg, {
			'grids': self.parent.grids.keys(),
			'updates': res,
		})
		return

	# -- FarmLink.message__queryinfo
	def message__queryinfo(self, msg):
		params = msg
		res = {}

		if 'grids' in params:
			res['grids'] = self.parent.grids.keys()

		if 'nodes' in params:
			if type(params['nodes']) != list:
				gnames = self.parent.grids.keys()
			else:
				gnames = params['nodes']

			nodes = {}
			grids = self.parent.grids
			for gname in gnames:
				if gname not in grids:
					nodes[gname] = None
					continue
				nodes[gname] = {
					'register': grids[gname].info['nodelist'],
					'online':  grids[gname].info['meter'].keys(),
				}
			res['nodes'] = nodes

		self.replyMessageWithResult(msg, res)
		return
	
	# -- FarmLink.message__executescript
	def message__executescript(self, msg):
		local_ns = {'PARAMS': msg.get('params')}
		global_ns = {
			'C': og.constants, 
			'link': self,
			'root': self.parent,
		}
		execption = None

		try:
			eval(msg.get('code'), local_ns, global_ns)
		except Exception, e:
			execption = e

		remsg = MessageReply(msg)
		if execption:
			res = {'exception': repr(e) }
		else:
			if 'RESULT' in local_ns:
				res = {'result': local_ns['RESULT']}
			elif 'RESULT' in global_ns:
				res = {'result': global_ns['RESULT']}
			else:
				res = {'result': 'OK'}

		self.replyMessageWithResult(msg, res)

# -- End of FarmLink
			

# -- Tasklet
# interface Tasklet
#   inMessage
#   exectption
#   timeout
#   close

# -- Tasklet
class TaskletInterface:
	# -- TaskletInterface.fire
	def fire(self): 
		TaskMap[self.name] = self
		self.start()

	def clean(self):
		del TaskMap[self.name]
		self.finished = True

	def close(self):
		self.clean()

	def onMessageIn(self, msg): pass
	def onExecption(self, why): self.close()
	def onTimeout(self): self.close()

# -- GridUpdateInfoTasklet
class GridUpdateInfoTasklet(TaskletInterface):
	# -- GridUpdateInfoTasklet.__init__
	def __init__(self, parent):
		self.finished = False
		self.name = '%s-grid-startup-%d' % (
			parent.name, Serial())
		self.last = Current.I
		self.timeout = 30
		self.parent = parent
		self.info = {}

	def messageOut(self, msg):
		self.parent.link.messageOut(msg)

	def start(self):
		self.tagline = 'get version',
		self.messageOut({
			'@': {'=': self.name},
			'!': {':': 'hive'},
			'#': Serial(),
			'_': 'S.versioninfo',
		})

	def onMessageIn(self, msg):
		self.last = Current.I
		if msg['_'] == 'S.versioninfo':
			self.info['version'] = msg['=']['version']
			self.messageOut({
				'@': {'=': self.name},
				'!': {'.':'hive', ':':'nodeinfo'},
				'#': Serial(),
				'_': 'S.nodes',
			})
			self.tagline = 'get node list in nodeinfo db'
		if msg['_'] == 'S.nodes':
			self.info['nodelist'] = msg['=']
			self.messageOut({
				'@': {'=': self.name},
				'!': {':': 'hive'},
				'#': Serial(),
				'_': 'S.nodemeterlist',
			})
			self.tagline = 'get nodemeterlist'

		elif msg['_'] == 'S.nodemeterlist':
			self.info['meter'] = msg['=']
			self.parent.info.update(self.info)
			self.parent.last['gridinfo'] = Current.I
			#_L('grid info %s==%s, %d nodes/%d online' % (
			#self.parent.name, self.info['version'], len(self.info['nodelist']),
			#len(self.info['meter'])))
			self.close()
	

# -- GridStartupUpdateInfoTasklet
class GridStartupUpdateInfoTasklet(GridUpdateInfoTasklet):
	def close(self):
		_L('Connect to grid %s==%s, %d nodes/%d online' % (
			self.parent.name, self.info['version'], len(self.info['nodelist']),
			len(self.info['meter'])))
		GridUpdateInfoTasklet.close(self)

# --
def Main():
	_C = og.constants
	def _E(name):
		name = 'OPENGRID_%s' % name.upper()
		return os.environ.get(name, getattr(_C, name))

	bp = {
		'home': _E('home'),
		'nodeid': 'farm',
	}

	os.chdir(bp['home'])
	logCore.setNodeId('FARM')

	cfname = '%s/etc/farm.cf' % bp['home']
	C = LoadConfig3(cfname, {'_C': _C}, None)
	C['boot-param'] = bp
	C['cfname'] = cfname

	F = Farming()
	F.setup(C)
	F.run()
# -- END --
