# ==
# --Project: Open Grid
# --Author: Sing.Yu <scaner@gmail.com>
# --Module: og.svc.hiving
# ==

HIVE_VERSION = '0.2-r26'

# ==
# hiving: Open Grid node hub service
# ==

import os, sys
import time

# --
import og.constants
import og.versioninfo

from og.lib.async import Poller, SocketDispatcher, OGConnection
from og.lib.async import TCPIncoming, MessageChannel

# --
from og.lib.logging import *
from og.lib.logging import logCore

# --
from og.lib.utility import SecondTimer, ListenAddr, MessageReply, SetUidAndGid, Serial
from og.lib.nodeinfo import NodeInfoRepository

# ==
# class Node
# class NodeLink
# class Hiving
# --

# --
INTERVAL_TOUCH_NODES = 2
INTERVAL_PING_NODES = 10

# == == == == == == == == == == == == == == == == == == == == == == #
# -- Node
class Node:
	# -- Node.__init__
	def __init__(self, parent, nodeid, **P):
		self.parent = parent 
		self.nodeid = nodeid
		self.startid = 0
		self.link = None
		self.info = {}
		self.omsg = None
		self.omsgQueue = []
		self.meter = None

		if P:
			self.setInfo(P)

		self.last = {
			'up': Current.I,
			'pong': Current.I,
			'meter': Current.I
		}

	# -- Node.info
	def setInfo(self, P = None, **F): 
		if P:
			self.info.update(P)
		else:
			self.info.update(F)

	# -- Node.touch
	def touch(self, curI):
		if not self.link and curI - self.last['up'] > 10:
			self.parent.notify('node.down.link', self)
			return
		if curI - self.last['meter'] > 70:
			self.parent.notify('node.down.meter', self)
			if self.link:
				self.link.close()
			return

	# -- Node.messageOut
	def messageOut(self, msg):
		if self.omsg or not self.link:
			self.omsgQueue.append(msg)
		else:
			self.omsg = msg
			self.link.messageOut(msg)
	
	# -- Node.onMessageOut
	def onMessageOut(self):
		self.omsg = None
		if not self.omsgQueue:
			return
		msg = self.omsgQueue.pop(0)
		self.omsg = msg
		self.link.messageOut(msg)

	# -- Node.onStartIDChanged
	def onStartIDChanged(self, newid):
		self.parent.notify('node.startidchanged', self)
		self.startid = newid
	
	# -- Node.onLinkUp
	def onLinkUp(self):
		_D('Node link from .%s established' % (self.nodeid))
		self.last['up'] = Current.I
		self.link.messageOut({
			'_': 'L.himain',
			'version': HIVE_VERSION,
			'protocol': 1,
		})
# -- End Of Node


# -- NodeLink
class NodeLink(object):
	# -- NodeLink.__init__
	def __init__(self, node):
		self.node = node
		self.routing = node.parent.routing
		self.C = None
	
	# -- NodeLink.setConnection
	def setConnection(self, conn):
		self.C = conn

	# -- NodeLink.onClose
	def onClose(self):
		if self.node.link == self:
			self.node.last['up'] = Current.I
			self.node.link = None

	# -- NodeLink.onMessageOut
	def onMessageOut(self):
		self.node.onMessageOut()

	# -- NodeLink.onMessageIn
	def onMessageIn(self, msg):
		self.routing(msg)
	
	# -- NodeLink.messageOut
	def messageOut(self, msg):
		self.C.sendMessage(msg)
	
	# -- NodeLink.onMessageOut
	def onMessageOut(self):
		self.node.onMessageOut()
	
	# -- NodeLink.close
	def close(self):
		self.C.onClose()
# -- End of NodeLink


# -- InfoLink
class InfoLink(MessageChannel):
	# -- InfoLink.__init__
	def __init__(self, parent):
		MessageChannel.__init__(self)
		self.parent = parent
		self.id = Serial()

		self.msgaddr = {
			'.': 'hive',
			':': 'infolink',
			'/': self.id,
		}
	
	# -- InfoLink.onConnect
	def onConnect(self):
		self.parent.infolinks[self.id] = self
	
	# -- InfoLink.onClose
	def onClose(self):
		del self.parent.infolinks[self.id]

	# -- InfoLink.onMessageIn
	def onMessageIn(self, msg):
		if msg['_'] == 'L.hihive':
			self.onHiHive(msg)
			return

		if '@' in msg:
			msg['@'].update(self.msgaddr)
		else:
			msg['@'] = self.msgaddr
		return self.parent.routing(msg)
	
	# -- InfoLink.onHiHive
	def onHiHive(self, msg):
		nodeid = msg['nodeid']
		startid = msg['startid']

		if msg.get('pin') != self.parent.nodeinfo.get(nodeid, 'pin'):
			_D('Node link error from .%s invalid pin' % nodeid)
			self.close()
			return

		# -- parameters: startid, version, protocol
		node = self.parent.getNode(nodeid)

		if not node:
			_D('Node link error from .%s unknown node id' % nodeid)
			self.close()
			return

		# TODO Check Link
		if node.startid != msg['startid']:
			node.setInfo(protocol = msg['protocol'], version = msg['version'])

		# clean Old node link if nessesry
		if node.link:
			node.link.close()

		node.link = self.toNodeLink(node)
		node.setInfo(addr = node.link.C.socket.getpeername())
		node.onLinkUp()
		if node.startid != startid:
			node.onStartIDChanged(startid)
	
	# -- LinkInfo.toNodeLink
	def toNodeLink(self, node):
		link = NodeLink(node)
		self.C.setHandler(link)
		self.onClose() # Clean reference in hiving.infolinks
		return link
# -- End of InfoLink

# == == == == == == == == == == == == == == == == == == == == == == #

# ==
# Hiving: hiving service main class
# __init__:
#   poller - Poller
#   alive - flag for running status
#   timer - Second timer
#   serial - Int serial generate
# setup:
#   ss - HivingLinkServer
#   lastCheck
# 
# ==
# -- Hiving
class Hiving:
	# -- Hiving.__init__
	def __init__(self):
		self.poller = Poller()
		self.poller.check = self.check
		self.alive = True
		self.nodes = {}
		self.infolinks = {}
		self.timer = SecondTimer(self)

		UpdateCurrent()
		self.startid = Current.I
		self.nodeid = 'hive'
	
	# -- Hiving.setup
	def setup(self, C):
		self.config = C
		self.nodeid = C['boot-param']['nodeid']
		
		self.nodeinfo = NodeInfoRepository(self)
		try:
			self.nodeinfo.loads()
		except:
			pass

		# create server socket
		self.ss = TCPIncoming(self, InfoLink)
		self.ss.initialize(ListenAddr(None, og.constants.HIVING_PORT))

		# setup timer
		self.lastCheck = Current.I
		self.timer.addCallback('nodes', Hiving.checkNodes, INTERVAL_TOUCH_NODES)
		self.timer.addCallback('pings', Hiving.pingNodes, INTERVAL_PING_NODES)

	# -- Hiving.check
	def check(self):
		UpdateCurrent()
		curI = Current.I

		if curI == self.lastCheck:
			return
			
		self.lastCheck = curI
		self.timer.touch(curI)
	
	# -- Hiving.pingNodes
	def pingNodes(self, cur):
		for n in self.nodes.keys():
			node = self.nodes[n]
			if not node.link:
				continue

			self.nodes[n].messageOut({
				'_': 'S.ping',
				'@': {'.': 'hive', ':': 'hive'},
				'!': {'.': n, ':':'main'},
				'$': 'NORETURN',
				'current': cur,
			})

	# -- Hiving.checkNodes
	def checkNodes(self, curI):
		for n in self.nodes.keys():
			self.nodes[n].touch(curI)
				
	# -- Hiving.RESTARTSERVICE
	def RESTARTSERVICE(self):
		for i in range(3, 1024):
			try: os.close(i)
			except OSError: pass
		os.execlp('ogstart', 'ogstart', 'hiving')
	
	# -- Hiving.getNode
	def getNode(self, nodeid):
		node = self.nodes.get(nodeid)
		if node:
			return node

		if nodeid not in self.nodeinfo.nodes():
			return None

		node = Node(self, nodeid)
		self.nodes[nodeid] = node
		return node

	# -- Hiving.notify
	def notify(self, msg, obj):
		if msg.startswith('node.down'):
			node = obj
			_L('Drop node info .%s' % node.nodeid)
			del self.nodes[node.nodeid]
			return

	# ==== Hiving Message handler ====
	# -- Hiving.routing
	def routing(self, msg):
		dest = msg['!']

		# To hive service
		if dest[':'] == 'hive':
			self.onMessage(msg)
			return

		# Hive info link
		if dest['.'] == 'hive':
			if dest[':'] == 'infolink':
				link = self.infolinks.get(dest['/'])
				if link == None:
					self.errorMessage(msg)
					return
				link.messageOut(msg)
				return
			elif dest[':'] == 'nodeinfo':
				self.nodeinfo.onMessage(msg)
				return 
			else:
				self.errorMessage(msg)
			return

		# To nodes
		node = self.nodes.get(dest['.'])
		if node:
			node.messageOut(msg)
			return
		self.errorMessage(msg)

	# -- Hiving.errorMessage
	def errorMessage(self, msg, why = 'S.ROUTING', where = 'anywhere'):
		# don't reply or relay error message
		if '$' in msg:
			return

		msg['!'] = msg['@']
		msg['$'] = (why, where)
		msg['@'] = {'.': 'hive', ':': 'core'}
		self.routing(msg)

	# -- Hiving.onMessage
	def onMessage(self, msg):
		fname = 'message__' + msg['_'][2:]
		f = getattr(Hiving, fname, Hiving.message__misc)
		f(self, msg)

	# -- Hiving.message__misc
	def message__misc(self, msg):
		_D('Misc message %s' % msg['_'])
	
	# -- Hiving.message__echo
	def message__echo(self, msg):
		reply = MessageReply(msg)
		msg.update(reply)
		msg['='] = 'OK'
		self.routing(msg)

	# -- Hiving.message__nodelist
	def message__nodelist(self, msg):
		res = {}
		nodes = self.nodes

		for n in nodes.values():
			res[n.nodeid] = {
				'nodeid': n.nodeid,
				'version': n.info['version'],
				'startid': n.startid,
				'online': n.link != None,
				'addr': n.info['addr'],
			}

		reply = MessageReply(msg)
		reply['='] = res
		self.routing(reply)
	
	# -- Hiving.message__nodemeterlist
	def message__nodemeterlist(self, msg):
		res = {}
		nodes = self.nodes

		for n in nodes.values():
			res[n.nodeid] = {
				'nodeid': n.nodeid,
				'version': n.info['version'],
				'startid': n.startid,
				'online': n.link != None,
				'addr': n.info['addr'],
				'meter': n.meter,
			}

		reply = MessageReply(msg)
		reply['='] = res
		self.routing(reply)
	
	# -- Hiving.message__pong
	def message__pong(self, msg):
		nodeid = msg['@']['.']
		n = self.nodes.get(nodeid)
		if not n:
			return
		n.last['pong'] = Current.I
	
	# -- Hiving.message__versioninfo
	def message__versioninfo(self, msg):
		remsg = MessageReply(msg)
		remsg['='] = {
			'version': HIVE_VERSION,
			'startid': self.startid,
		}
		self.routing(remsg)
	
	# -- Hiving.message__sourcerevision
	def message__sourcerevision(self, msg):
		remsg = MessageReply(msg)
		remsg['='] = {
			'version-info': og.versioninfo.version_info,
			'file-revisions': og.versioninfo.file_revisions,
		}
		self.routing(remsg)

	# -- Hiving.message__meter
	def message__meter(self, msg):
		nodeid = msg['@']['.']
		n = self.nodes.get(nodeid)
		if not n:
			return
		n.meter = msg['=']
		n.last['meter'] = Current.I

	# -- Hiving.message__executescript
	def message__executescript(self, msg):
		local_ns = {'PARAMS': msg.get('params')}
		global_ns = {'C': og.constants, 'root': self}
		execption = None

		try:
			eval(msg.get('code'), local_ns, global_ns)
		except Exception, e:
			execption = e

		remsg = MessageReply(msg)
		if execption:
			remsg['='] = {'exception': repr(e) }
		else:
			if 'RESULT' in local_ns:
				remsg['='] = {'result': local_ns['RESULT']}
			elif 'RESULT' in global_ns:
				remsg['='] = {'result': global_ns['RESULT']}
			else:
				remsg['='] = {'result': 'OK'}
		self.routing(remsg)

	# -- Hiving.message__restartservice
	def message__restartservice(self, msg):
		self.RESTARTSERVICE()

	# ==== main running loop ====
	# -- Hiving.run
	def run(self):
		_I('Hive-%s:%d ready' % (HIVE_VERSION, os.getpid()))
		self.poller.mainloop()
# -- End of Hiving

# -- Main
def Main():
	_C = og.constants
	_E = os.environ
	def __E(name):
		keyname = 'OPENGRID_%s' % (name.upper())
		return _E.get(keyname, getattr(_C, keyname))

	bp = dict([(x, __E(x)) for x in ('domain', 'home', 'nodeid', 'name')])
	os.chdir(bp['home'])

	# -- setup logCore
	logCore.setNodeId(bp['nodeid'], 'HIVE')
	logCore.setUDPAddr(('ulog--' + bp['domain'], _C.ULOG_PORT))

	# --
	H = Hiving()
	H.setup({'boot-param':bp})
	H.run()

# --
if __name__ == '__main__':
	Main()
