# ==
# --Project: Open Grid
# --Author: Sing Yu <scaner@gmail.com>
# --Module: og.svc.managing
# ==
MANAGING_VERSION = '0.2-r24'

# ==
# Managing Service:
#   Open Grid node managing service
# ==

# ---
import os, sys
import signal
import time

# ---
from errno import EWOULDBLOCK 

# ---
from og.lib.utility import ListenAddr, FirstExists, LoadConfig3
from og.lib.utility import SecondTimer, SetUidAndGid, Serial
from og.lib.utility import MessageReply, MessageError

from og.lib.logging import *
from og.lib.logging import logCore

from og.lib.async import Poller, TCPIncoming
from og.lib.async import OGConnection, MessageChannel
from og.lib.async import DispatcherBase

from og.lib.process import ProcInfo, ProcManager
from og.lib.egg import EggRotate

import og.constants
import og.versioninfo

import og.lib.osmeter

# == == == == == == == == == == == == == == == == == == == == == == #
INTERVAL_HIVE_LINK_CHECK = 2
INTERVAL_METER_UPDATE = 45

# == == == == == == == == == == == == == == == == == == == == == == #
# -- HiveLink
class HiveLink(MessageChannel):
	# -- HiveLink.__init__
	def __init__(self, parent, hiveAddr):
		MessageChannel.__init__(self)
		self.parent = parent
		self.hiveAddr = hiveAddr
	
	# -- HiveLink.isActive
	def isActive(self):
		return self.C != None

	# -- HiveLink.connect
	def connect(self):
		c = OGConnection(self.parent.poller)
		c.createTCPSocket()
		c.setHandler(self)
		self.C.connect(self.hiveAddr)

	# -- HiveLink.onConnect
	def onConnect(self):
		self.C.sendMessage({
			'_': 'L.hihive',
			'pin': self.parent.myconfig['pin'],
			'nodeid': self.parent.nodeid,
			'startid': self.parent.startid,
			'version': MANAGING_VERSION,
			'protocol': 1
		})

	# -- HiveLink.onMessageIn
	def onMessageIn(self, msg):
		if msg['_'] == 'L.himain':
			_D('Hive link established')
		elif msg['_'] == 'L.restart' or msg['_'] == 'RESTARTSERVICE':
			self.parent.RESTARTSERVICE()
		else:
			self.parent.routing(msg)
	
	# -- HiveLink.onClose
	def onClose(self):
		self.C = None
		
		# push sending message out
		if self.oMessage:
			self.oMessages.insert(0, self.oMessage)
			self.oMessage = None
# -- End of HiveLink
	

# == == == == == == == == == == == == == == == == == == == == == == #
# -- Managing
class Managing:
	# -- Managing.__init__
	def __init__(self):
		self.pm = ProcManager(self)
		self.hasZombie = False
		self.alive = True

		self.poller = Poller()
		self.poller.check = self.check

		self.tasklets = {}
		self.meter = og.lib.osmeter.OSMeter()

	# -- Managing.sigChildren
	def sigChildren(self, signo, frame):
		self.hasZombie = True
	
	# -- Managing.setup
	def setup(self, C):
		self.config = C
		self.myconfig = C.get('managing')

		self.nodeid = C['nodeid']

		C = self.myconfig
		if not C:
			C = {}

		# create server socket.
		self.ss = TCPIncoming(self, HiveLink)
		addr = ListenAddr(C.get('listen'), 2345)
		self.ss.initialize(addr)
		
		# register signal handler
		signal.signal(signal.SIGCHLD, self.sigChildren)

		# last touch
		UpdateCurrent()
		self.lastCheck = 0
		self.startid = Current.I

		# Setup
		self.setupEnvironment()
		self.startupProcess()

		# Setup timer
		self.timer = SecondTimer(self)
		self.timer.addCallback('hivelink', 
			Managing.checkHiveLink, INTERVAL_HIVE_LINK_CHECK)
		self.timer.addCallback('meter', 
			Managing.updateMeter, INTERVAL_METER_UPDATE)
			

		# Setup Link to Hive
		self.hiveLink = HiveLink(self, (
			'hive--' + self.config['boot-param']['domain'], 
			og.constants.HIVING_PORT))
		self.hiveLink.connect()
	
	# -- Managing.startupProcess
	def startupProcess(self):
		C = self.myconfig.get('startup')
		if not C:
			return

		for param in C:
			pi = ProcInfo(self, **param)
			pi.fire()

	# -- Managing.setupEnvironment
	def setupEnvironment(self):
		# setup environment
		env = os.environ
		bootParam = self.config['boot-param']

		env['OPENGRID_HOME'] = bootParam['home']
		env['OPENGRID_DOMAIN'] = bootParam['domain']
		env['OPENGRID_NAME'] = bootParam['name']
		env['OPENGRID_NODEID'] = self.config['nodeid']


	# -- Managing.run
	def run(self):
		_I('Managing-%s:%d ready' % (MANAGING_VERSION, os.getpid()))
		self.poller.mainloop()
		
	# -- Managing.updateMeter
	def updateMeter(self, cur):
		self.meter.touch()
		self.meterNotify()
	
	# -- Managing.meterNotify
	def meterNotify(self):
		meterinfo = self.meter.info()
		meterinfo['current'] = Current.I

		self.routing({
			'_': 'S.meter',
			'!': {'.': 'hive', ':':'hive'},
			'@': {'.': self.nodeid, ':':'main'},
			'$': 'no-error-return',
			'#': Serial(),
			'=': meterinfo,
		})
		
	# -- Managing.checkHiveLink:
	def checkHiveLink(self, cur):
		if not self.hiveLink.isActive():
			self.hiveLink.connect()

	# -- Managing.check
	def check(self):
		if self.hasZombie:
			self.hasZombie = False
			self.pm.reapZombie()
	
		UpdateCurrent()
		curI = Current.I
		if curI == self.lastCheck:
			return
		self.lastCheck = curI
		self.timer.touch(curI)


	# -- Managing.RESTARTSERVICE
	def RESTARTSERVICE(self):
		# disable alive flag
		self.alive = False

		self.pm.killall()

		# close all file descriptor beside 0, 1, 2
		for i in range(3, 1024):
			try:
				os.close(i)
			except OSError:
				pass

		# --
		# pid = os.fork()
		# if pid != 0:
		#	os.exit()
		# --
		os.execlp('ogstart', 'ogstart', 'maining')

	# -- Managing.routing
	def routing(self, msg):
		dest = msg['!']

		nodeid = dest.get('.', self.nodeid)
		if nodeid == self.nodeid:
			# TODO Local message dispatcher
			self.onMessage(msg)
			return

		# message to other node will be send hiving directly
		self.hiveLink.messageOut(msg)

	# -- Managing.errorMessage
	def errorMessage(self, msg, why = 'S.ROUTING', where = 'anywhere'):
		if '$' in msg:
			return

		remsg = MessageError(msg,
			{'.': self.nodeid, ':':'main'},
			why, where)
		self.routing(remsg)

	# -- Managing.onMessage
	def onMessage(self, msg):
		func = getattr(Managing, 'message__' + msg['_'][2:], Managing.message__misc)
		func(self, msg)

	# -- Managing.message__misc
	def message__misc(self, msg):
		_D('Misc message %s' % msg['_'])

	# -- Managing.message__echo
	def message__echo(self, msg):
		remsg = MessageReply(msg)
		msg.update(remsg)
		msg['='] = 'OK'
		self.routing(msg)
	
	# -- Managing.message__ping
	def message__ping(self, msg):
		remsg = MessageReply(msg)
		remsg['_'] = 'S.pong'
		self.routing(remsg)

	# -- Managing.message__restart
	def message__restart(self, msg):
		self.RESTARTSERVICE()
	
	# -- Managing.message__set
	def message__set(self, msg):
		params = msg['params']
		for k, v in params.items():
			if k == 'meter.interval':
				self.timer.setInterval('meter', v)

		remsg = MessageReply(msg)
		remsg['=']= 'OK'
		self.routing(remsg)
		
	# -- Managing.message__run0
	def message__run0(self, msg):
		tasklet = SimpleProcTask(self)
		tasklet.setup(msg)
		tasklet.fire()
		remsg = MessageReply(msg)
		remsg['='] = {'state':'fire', 'pid': tasklet.pid}
		self.routing(remsg)
	
	# -- Managing.message__versioninfo
	def message__versioninfo(self, msg):
		remsg = MessageReply(msg)
		res = {
			'version': MANAGING_VERSION,
			'startid': self.startid,
		}
		remsg['='] = res
		self.routing(remsg)

	# -- Managing.message__sourcerevision
	def message__sourcerevision(self, msg):
		remsg = MessageReply(msg)
		remsg['='] = {
			'version-info': og.versioninfo.version_info,
			'file-revisions': og.versionfino.file_revisions,
		}
		self.routing(remsg)

	# -- Managing.message__cleaneggs
	def message__cleaneggs(self, msg):
		remsg = MessageReply(msg)
		remsg['='] = {}
		for e in ('opengrid', 'fastmap'):
			remsg['='][e] = EggRotate(
				os.environ['OPENGRID_HOME'] + '/lib/ogeggs', 
				e + '-')
		self.routing(remsg)

# -- End of Managing

# == == == == == == == == == == == == == == == == == == == == == == #
class SimpleProcTaskReadStream(DispatcherBase):
	# -- SimpleProcTaskReadStream.__init__
	def __init__(self, poller, fd):
		DispatcherBase.__init__(self, poller)
		self._fileno = fd
		self.closed = False
		self.buffer = None
		self.attachChannel()
	
	# -- SimpleProcTaskReadStream.onEventRead
	def onEventRead(self):
		while True:
			try:
				buffer = os.read(self._fileno, 4096)
				if not buffer:
					self.close()
					return
				if self.buffer != None:
					self.buffer.append(buffer)
			except OSError, why:
				if why[0] != EWOULDBLOCK:
					raise
				return
		
	# -- SimpleProcTaskReadStream.onEventException
	def onEventException(self):
		self.close()

	# -- SimpleProcTaskReadStream.setBuffer
	def setBuffer(self, buffer):
		self.buffer = buffer
	
	# -- SimpleProcTaskReadStream.enable
	def enable(self):
		self.registerRead(True)
	
	# -- SimpleProcTaskReadStream.close
	def close(self):
		if self.closed: return

		self.registerRead(False)
		self.registerWrite(False)
		self.detachChannel()
		os.close(self._fileno)
		self.closed = True
# -- End of SimpleProcTaskReadStream

class SimpleProcTaskWriteStream(DispatcherBase):
	# -- SimpleProcTaskWriteStream.__init__
	def __init__(self, poller, fd):
		DispatcherBase.__init__(self, poller)
		self._fileno = fd
		self.closed = False
		self.keepOpen = False
		self.buffer = None
		self.attachChannel()
	
	# -- SimpleProcTaskWriteStream.setKeepOpen
	def setKeepOpen(self, keep):
		self.keepOpen = keep

	# -- SimpleProcTaskWriteStream.setBuffer
	def setBuffer(self, buffer):
		self.buffer = buffer
	
	# -- SimpleProcTaskWriteStream.enable
	def enable(self):
		if self.buffer:
			self.registerWrite(True)
		else:
			if self.keepOpen:
				self.registerRead(True)
			else:
				self.close()
	
	# -- SimpleProcTaskWriteStream.onEventWrite
	def onEventWrite(self):
		while self.buffer:
			try:
				ret = os.write(self._fileno, self.buffer)
			except OSError, why:
				if why[0] != EWOULDBLOCK:
					raise
				return

			if ret == len(self.buffer):
				self.buffer = None
			else:
				self.buffer = self.buffer[ret:]
		
		if self.keepOpen:
			self.registerWrite(False)
			self.registerRead(True)
		else:
			self.close()

	# -- SimpleProcTaskWriteStream.onEventException
	def onEventException(self):
		self.close()
	
	# -- SimpleProcTaskWriteStream.close
	def close(self):
		if self.closed: return

		self.registerWrite(False)
		self.registerRead(False)
		self.detachChannel()
		os.close(self._fileno)
		self.closed = True

# -- End of SimpleProcTaskWriteStream
# ==
# S.run0
#   args: arguments
#   env: envrionments
#   stdin: stdin text
#   stdout: record stdout
#   stderr: record stderr
#   name: name & id
#   '~'/tag: tag
# ==
class SimpleProcTask:
	# -- SimpleProcTask.__init__
	def __init__(self, managing):
		self.managing = managing
	
	# -- SimpleProcTask.setup
	def setup(self, msg):
		self.pi = ProcInfo(self.managing.pm,
			args = msg.get('args'),
			env = msg.get('env'),
			name = msg.get('name'),
			flags = 'P',
		)

		self.stdinKeepOpen = msg.get('stdin-keep-open', False)
		self.msgfrom = msg['@']
		self.msgtag = msg.get('~')

		self.stdin = msg.get('stdin')
		self.stdout = msg.get('stdout')
		self.stderr = msg.get('stderr')

		self.pi.setNotifier(self)
		self.managing.tasklets[self.msgtag] = self

		
	# -- SimpleProcTask.fire
	def fire(self):
		self.pi.fire()

		self.fd0 = SimpleProcTaskWriteStream(self.managing.poller, self.pi.stdin)
		self.fd1 = SimpleProcTaskReadStream(self.managing.poller, self.pi.stdout)
		self.fd2 = SimpleProcTaskReadStream(self.managing.poller, self.pi.stderr)

		if self.stdinKeepOpen:
			self.fd0.setKeepOpen(True)
		if self.stdin: self.fd0.setBuffer(self.stdin)

		if self.stdout:
			self.stdoutBuffer = []
			self.fd1.setBuffer(self.stdoutBuffer)
		if self.stderr:
			self.stderrBuffer = []
			self.fd2.setBuffer(self.stderrBuffer)

		self.fd0.enable()
		self.fd1.enable()
		self.fd2.enable()

		self.pid = self.pi.pid
	
	# -- SimpleProcTask.onFork
	def onFork(self, PI, pid): pass
	# -- SimpleProcTask.onExecute
	def onExecute(self, PI): pass

	def onReap(self, PI, pid, status):
		if self.managing.tasklets.get(self.msgtag) == self:
			del self.managing.tasklets[self.msgtag]

		msg = {
			'_': 'S.run0.finished',
			'!': self.msgfrom,
			'@': {'.': self.managing.nodeid, ':':'main'},
			'~': self.msgtag,
			'#': Serial()
		}

		self.fd0.close()
		self.fd1.close()
		self.fd2.close()

		res = {
			'status': status,
			'pid': pid,
		}

		if self.stdout:
			res['stdout'] = ''.join(self.fd1.buffer)
		if self.stderr:
			res['stderr'] = ''.join(self.fd2.buffer)
		msg['='] = res
		self.managing.routing(msg)
# -- End of SimpleProcTask	

# ---
# TODO: empty autoconfigure
def OGAutoConfig(bc, cfname):
	return {}

# ---
def OGBootConfig(bootParam, cfname):
	C = LoadConfig3(cfname, {'BootParam':bootParam}, None)

	C['boot-param'] = bootParam
	C['cfname'] = cfname

	return C

# ---
def OGGetConfig(param):
	BP = os.path.join

	# -- Try auto configure
	base = param['home']
	name = param['name']
	p0 = BP(base, 'etc')
	p1 = BP('/etc', 'opengrid')

	autobootcfnames = (
		BP(p0, 'autoconf.%s' % name),
		BP(p0, 'autoconf'),
		BP(p1, 'autoconf.%s' % name),
		BP(p1, 'autoconf'))

	autobootcfname = FirstExists(*autobootcfnames)

	if autobootcfname:
		return OGAutoConfig(param, autobootcfname)
	
	# -- Read configure from boot.conf
	bootcfnames = (
		BP(p0, 'main.%s.cf' % name),
		BP(p0, 'main.cf'),
		BP(p1, 'main.%s.cf' % name),
		BP(p1, 'main.cf'))
	bootcfname = FirstExists(*bootcfnames)

	if bootcfname:
		return OGBootConfig(param, bootcfname)
	raise Exception, "Can not locate boot configure"
	
# ---
def FastConfig0(name, default = None):
	if not os.path.exists(name):
		return default
	fin = file(name,'r')
	value = fin.read().strip()
	fin.close()
	return value

# ---
def FastConfig(home, name, default):
	return FastConfig0(home + '/etc/boot-params/' + name, default)
	

# ---
def FastStart():
	_C = og.constants

	home = os.environ.get('OPENGRID_HOME', _C.OPENGRID_HOME)
	nodeid = FastConfig(home, 'nodeid', _C.OPENGRID_NODEID)
	domain = FastConfig(home, 'domain', _C.OPENGRID_DOMAIN)
	name = FastConfig(home, 'gridname', _C.OPENGRID_NAME)

	bootParam = {
		'domain': domain,
		'name': name,
		'nodeid': nodeid,
		'home': home,
	}

	C = {
		'managing': {
			'listen': ('127.0.0.1', _C.MANAGING_PORT),
			'pin': FastConfig(home, 'pin', 'XXXXXXXXXXXXXXXX')
		},
		'nodeid': nodeid,
		'boot-param': bootParam
	}

	logCore.setup(C, 'main')
	M = Managing()
	M.setup(C)
	M.run()

# ---
def Main():
	SetUidAndGid()
	_C = og.constants

	bootParam = {
		'domain': os.environ.get('OPENGRID_DOMAIN', _C.OPENGRID_DOMAIN),
		'name': os.environ.get('OPENGRID_NAME', _C.OPENGRID_NAME),
		'home': os.environ.get('OPENGRID_HOME', _C.OPENGRID_HOME),
		'nodeid': os.environ.get('OPENGRID_NODEID', _C.OPENGRID_NODEID),
	}

	os.chdir(bootParam['home'])
	C = OGGetConfig(bootParam)
	logCore.setup(C, 'main')
	_I('Use %s as boot-conf' % C['cfname'])

	M = Managing()
	M.setup(C)
	M.run()

# ---
if __name__ == '__main__':
	Main()
